package bsplines;

/*
Tutorial on LevelSet and BSpline deformable model techniques 
for image segmentation
Copyright (C) Diane Lingrand (Diane.Lingrand@unice.fr)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

import java.util.ArrayList;
import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.*;
import javax.imageio.*;
import java.io.*;
import java.util.jar.*;
import java.net.URL;
import java.net.MalformedURLException;

/**
 * BSPanel class :
 *    manages the evolution of a BSpline according to a BufferedImage
 * 
 * @author <a href="mailto:Diane.Lingrand@unice.fr">DianeLingrand</a>
 **/

public class BSPanel extends SegPanel implements MouseListener{
   
    private static final double DIFF_RATIO = 20.0;

    private ArrayList listBsp;
    private boolean inConstruction = true;
    private boolean withMask = false; // do we draw the mask ?

    private int close = BSpline.CLOSE;
    private int distant = BSpline.DISTANT;

    private int iter = 0;

    private BSpline bsc;

    public BSPanel(double l1, double l2, double l3) {
	this(IMAGE_FILENAME, l1, l2, l3);
    }

    public BSPanel(String filename, double l1, double l2, double l3) {
	super();

	lambda1 = l1;
	lambda2 = l2;
	lambda3 = l3;

	this.addMouseListener(this);
	setPreferredSize(new Dimension(w,h));
	setBackground(new Color(255,255,255));

	listBsp = new ArrayList();

	// opening an image in this panel
	URL imageURL = bsplines.BSApplet.class.getResource("images/"+filename);
	System.out.println("url = " + imageURL);
	if(imageURL != null) {
	    ImageIcon ii = new ImageIcon(imageURL);
	    bi = new BufferedImage(ii.getIconWidth(), ii.getIconHeight(), BufferedImage.TYPE_INT_RGB);
	    Graphics g = bi.getGraphics();
	    ii.paintIcon(this, g, 0, 0);
	}
	if(bi != null) {
	    w = bi.getWidth();
	    h = bi.getHeight();
	    biLiss = lissage(bi);
	    setPreferredSize(new Dimension(w,h));
	    inConstruction = false;
	    listBsp.add(new BSpline(BSpline.TORDU,w,h));
	}
	else{
	    System.out.println("Unable to open file " + filename);
 	    listBsp.add(new BSpline(BSpline.CIRCLE,200,200));
	}
    }

    public BSPanel(LSPanel pan) {
	super();

	lambda1 = pan.getLambda1();
	lambda2 = pan.getLambda2();
	lambda3 = pan.getLambda3();
	bi = pan.getBufferedImage();
	biLiss = pan.getLissBufferedImage();
	avecLissage = pan.getSmoothing();
	capture = pan.getCapture();

	double [][] ls = pan.getLevelSet();

	this.addMouseListener(this);
	setPreferredSize(new Dimension(220,220));
	setBackground(new Color(255,255,255));

	listBsp = new ArrayList();

	w = bi.getWidth();
	h = bi.getHeight();
	setPreferredSize(new Dimension(w,h));
	inConstruction = false;

	// we now constructs the BSplines corresponding to the levelSet
	// a better idea should be to use Marching Squares

	/// first step : we mark contours points
	boolean [][] contours = new boolean[w][h];
	for(int i = 0; i < w-1; i++)
	    for(int j = 0; j < h-1 ; j++) {
		int nbPos = 0;
		int nbNeg = 0;

		
		if( ls[i][j] > 0) nbPos++; else if( ls[i][j] <= 0) nbNeg++;
		if( ls[i+1][j] > 0) nbPos++; else if( ls[i+1][j] <= 0) nbNeg++;
		if( ls[i+1][j+1] > 0) nbPos++; else if( ls[i+1][j+1] <= 0) nbNeg++;
		if( ls[i][j+1] > 0) nbPos++; else if( ls[i][j+1] <= 0) nbNeg++;
		if( ( nbNeg < 4) && (nbPos < 4) )
		    contours[i][j] = true;
		else
		    contours[i][j] = false;
	    }
	/// second step : we chains points
		boolean chainFini = false;
		while(!chainFini) {
		    chainFini = true;
		    BSpline bsp = new BSpline();
		    int i = 0;
		    int j = 0;
		    boolean notFound = true;
		    for(i = 0; (i < w-2)&&(notFound); i++)
			for(j = 0 ; (j < h-2)&&(notFound) ; j++) {
			    if(contours[i][j]) {
				notFound = false;
				contours[i][j] = false;
				chainFini = false;
				// first point of the BSpline
				BSPoint p = new BSPoint(i,j);
				bsp.add(p);
			    }
			}
		    boolean fini = false;

		    while(!fini) {
			if((i < w-2) && contours[i+1][j]) { i++;}
			else if((j < h-2) && contours[i][j+1]) { j++; }
			else if((i > 0) && contours[i-1][j]) {	i--; }
			else if((j > 0) && contours[i][j-1]) { j--; }
 			else if( (i>0)&&(j>0)&& contours[i-1][j-1]) { i--; j--; }
 			else if( (i>0)&&(j<h-2)&& contours[i-1][j+1]) { i--; j++; }
 			else if( (i<w-2)&&(j>0)&& contours[i+1][j-1]) { i++; j--; }
 			else if( (i<w-2)&&(j<h-2)&& contours[i+1][j+1]) { i++; j++; }
			else fini = true;
			if(contours[i][j]) {
			    contours[i][j] = false;
			    BSPoint p = new BSPoint(i,j);
			    bsp.add(p);
			}
			
		    }
		    // here, we only take 1/10 points
		    BSpline newbsp = new BSpline();
		    for(int k = 0; k < bsp.size(); k+= 10)
			newbsp.add((BSPoint) bsp.get(k));
		    if(newbsp.size() > 3) {
			newbsp.computeQ();
			listBsp.add(newbsp);
		    }
		}
    }

    public ArrayList getListBsp() { return listBsp;}

    public void loadBSplines(BufferedReader reader) {
	try{
	    String str = reader.readLine();
	    int s = Integer.parseInt(str);
	    ArrayList newList = new ArrayList();
	    for(int i = 0; i < s; i++) {
		str = reader.readLine();
		int size = Integer.parseInt(str);
		BSpline bsp = new BSpline();
		for(int j = 0; j < size; j++) {
		    str = reader.readLine();
		    int b1 = str.indexOf(" ", 0);
		    BSPoint p = new BSPoint(Double.parseDouble(str.substring(0,b1)),Double.parseDouble(str.substring(b1+1,str.length())));
		    bsp.add(p);
		}
		bsp.computeQ();
		newList.add(bsp);
	    }
	    listBsp = newList;
	    reader.close();
	    repaint();
	}
	catch(IOException e) {
	    System.out.println("Unable to load file ");
	}
    }

    public void saveBSplines(File file) {
	try{
	    BufferedWriter writer = new BufferedWriter(new FileWriter(file));
	    writer.write(saveBSplinesAsString());
	    writer.flush();
	    writer.close();
	}
	catch(IOException e) {
	    System.out.println("Unable to write in file " + file);
	}
    }
    public String saveBSplinesAsString() {
	    int s = listBsp.size();
	    String ss = ""+ s +"\n";
	    
	    for(int i = 0; i < s; i++) {
		ss +=  ((BSpline) listBsp.get(i)).createString();
	    }
	    return ss;
    }
    public BSparam loadParameters(BufferedReader reader) {
	BSparam res = null;
	try{
	    String str = reader.readLine();
	    int b1 = str.indexOf(" ", 0);
	    res = new BSparam();
	    close = Integer.parseInt(str.substring(0,b1)); res.setClose(close);
	    distant = Integer.parseInt(str.substring(b1+1,str.length())); res.setDistant(distant);
	    str = reader.readLine();
	    criterion = Integer.parseInt(str); res.setCriterion(criterion);
	    str = reader.readLine();
	    b1 = str.indexOf(" ", 0);
	    int b2 = str.indexOf(" ",b1+1);
	    lambda1 = Double.parseDouble(str.substring(0,b1)); res.setLambda1(lambda1);
	    lambda2 = Double.parseDouble(str.substring(b1+1,b2)); res.setLambda2(lambda2);
	    lambda3 = Double.parseDouble(str.substring(b2+1,str.length())); res.setLambda3(lambda3);
	    System.out.println("close = " + close + " distant = " + distant);
	    System.out.println("criterion = " + criterion);
	    System.out.println("lambda1 = " + lambda1 + " lambda2 = " + lambda2 + " lambda3 = " + lambda3);
	    reader.close();

	}
	catch(IOException e) {
	    System.out.println("Unable to load file ");
	}
	return res;
    }
    /**
     * save the parameters in an ASCII file such as :
     * close distant
     * criterion
     * lambda1 lambda2 lambda3
     **/
    public void saveParameters(File file) {
	try{
	    BufferedWriter writer = new BufferedWriter(new FileWriter(file));

	    writer.write(""+ close + " " + distant + "\n");
	    writer.write(""+ criterion + "\n");
	    writer.write(""+ lambda1 + " " + lambda2 + " " + lambda3 + "\n");
	    writer.flush();
	    writer.close();
	}
	catch(IOException e) {
	    System.out.println("Unable to write in file " + file);
	}
    }
     public void saveParameters(JTextArea jt) {
	    jt.append(""+ close + " " + distant + "\n");
	    jt.append(""+ criterion + "\n");
	    jt.append(""+ lambda1 + " " + lambda2 + " " + lambda3 + "\n");
    }
    public void setClose(int c) {
	int s = listBsp.size();
	if( c > distant )
	    close = distant;
	else
	    close = c;
	for(int i = 0; i < s; i++)
	    ((BSpline) listBsp.get(i)).setClose(close);
    }

    public void setDistant(int d) {
	int s = listBsp.size();
	if( d < close )
	    distant = close;
	else
	    distant = d;
	for(int i = 0; i < s; i++)
	    ((BSpline) listBsp.get(i)).setDistant(distant);
    }
    
    /** 
     * clears all the Bsplines in this Panel
     **/
    public void clearBSplines() {
	bsc = new BSpline();
	bsc.setClose(close); bsc.setDistant(distant);
	listBsp = new ArrayList();
	listBsp.add(bsc);
	inConstruction = true;
	repaint();

    }
    public void changeDrawWithMask() {
	withMask = !withMask;
	repaint();
    }

    // implements ...
    public void mouseClicked(MouseEvent e){}
    public void mouseEntered(MouseEvent e){}
    public void mouseExited(MouseEvent e){}

    public void mousePressed(MouseEvent e){
	if(e.getButton() == MouseEvent.BUTTON2) {
	    changeDrawWithMask();
	    return;
	}

	if(inConstruction) {
	    if(e.getButton() == MouseEvent.BUTTON3) {
		inConstruction = false;
		bsc = null;
	    }
	    else {
		double x = (double) e.getX();
		double y = (double) e.getY();
		BSPoint p = new BSPoint(x,y);
		bsc.add(p);
	    }
	}
	else {
	    if(e.getButton() == MouseEvent.BUTTON3) {
		inConstruction = true;
		bsc = new BSpline();
		bsc.setClose(close);
		bsc.setDistant(distant);
		listBsp.add(bsc);
	    }
	    else
		evolve(true);
	}
    }

    public void mouseReleased(MouseEvent e){//evolve(false);
	if( e.getButton() == MouseEvent.BUTTON1)
	    if (inConstruction) {
		if(bsc.size() > 0)
		    bsc.computeQ();
		repaint();
	    }
	    else
		evolve(false);
    }

    /**
     * evoluting the BSpline according to image informations.
     **/
    public void evolve(boolean b) {
	int s = listBsp.size();

	if(s < 1) return;
	if(b && (bi == null)) {
	    for(int i = 0; i < s; i++) {
		BSpline bsp = (BSpline) listBsp.get(i);
		int size = bsp.size();
		if(size <=1 ) bsp.checkDistances();
		else bsp.evolve(0.01);
	    }
	    paintImmediately(0, 0, (int) getSize().getWidth(), (int) getSize().getHeight());
	    return;
	}

	// mouse down
	if(b) {
	    applyCrit();
	    cutBSplines();
	    reorientCurv();
	}
	// mouse up
	else {
	    regulDistances(0);
	}
	paintImmediately(0, 0, (int) getSize().getWidth(), (int) getSize().getHeight());// repaint();
    }

    /**
     * for each BSpline in the list of BSpline, examines if there is
     * self intersection. If yes, cut the BSpline into 2 parts
     **/

    public void cutBSplines() {
	
	for(int i = 0; i < listBsp.size(); i++) {
	    BSpline bsp = (BSpline) listBsp.get(i);
	    BSpline nbs;
	    do {
		nbs = bsp.cutBSpline(w, h);
		if(bsp.size() < 3){
		    listBsp.remove(i);
		    bsp = null;
		    i--;
		}
		if( (nbs != null) && (nbs.size() > 2)) {
		    System.out.println("adding new BSpline with size = " + nbs.size());
		    listBsp.add(nbs);
		}
	    } while((nbs != null)&&(bsp != null));
	}
    }

    /**
     * for each BSpline of the list of BSplines
     * if a Spline has less than 3 points, removes it
     * else, uniforms the distance between 2 consecutives BSPoints
     **/
    public void regulDistances(int var) {
	for(int i = 0; i < listBsp.size(); i++) {
	    BSpline bsp = (BSpline) listBsp.get(i);
	    if(bsp.size() < 3)
		listBsp.remove(i--);
	    else if(bsp.size() + var < 3)
		listBsp.remove(i--);
	    else
		bsp.regulDistances(var);
	}

	// draw and save the image with current bspline segmentation
	if(capture) {
	    BufferedImage bd = new BufferedImage(w,h,BufferedImage.TYPE_INT_RGB);
	    Graphics g = bd.getGraphics();
	    paintComponent(g);
	    try{
		if(iter<10)
		    ImageIO.write(bd,"png",new File("bspline0"+ iter +".png"));
		else
		    ImageIO.write(bd,"png",new File("bspline"+ iter +".png"));
	    }
	    catch(Exception e) {System.out.println("pas de regul bd");}
	    iter++;
	}
    }
  
    /**
     * for each BSpline, check if the orientation is OK (normals looking inside the object)
     * the simple algorithm is based that, often (not always), if we take 3 points P0, P1 and P2
     * along the curve, verifying that the arc distances P0P1, P1P2 and P2P0 are equals,
     * the sign of the cross product P0P1xP0p2 must be positive. If not, we inverse the points
     * order in the BSpline
     **/
    public void reorientCurv() {

	for(int i = 0; i < listBsp.size(); i++) {
	    BSpline bsp = (BSpline) listBsp.get(i);
	    if(bsp.size() < 3)
		listBsp.remove(i--);
	    else
		bsp.reorientCurv();
	}
    }


    /**
     * evolution criterion : (I-muIn)^2 - (I-muOut)^2 + curv
     * assuming that the object and the background are homogeneous
     **/
    public void evolveMean() {
	int s = listBsp.size();
	int [][] m;
	double muI, muO;
	double crit;
	int puissDeux = 1; //

	m = createMask2();

	muO = meanOutside(m);
	// for each BSpline
	for(int b = 0; b < s; b++) {
	    BSpline bsp = (BSpline) listBsp.get(b);
	    muI = meanInside(m,puissDeux); // muI = meanInside(m,b+1);
	    // for each BSPoint
	    int size = bsp.size();
	    for(int i = 0; i < size; i++) {
		BSPoint p = bsp.get(i);
		int x = (int) (p.getXP());
		int y = (int) (p.getYP());
		if( (x<0) || (y<0) || (x >= w)|| (y >= h) )
		    crit = 0.0;
		else {
		    int I = getI(x,y);
		    double curvTerm = p.getCurv();
		    if( curvTerm < 0) curvTerm = 0.0;
		    crit = lambda1 * (I-muI)*(I-muI) - lambda2 * (I-muO)*(I-muO) + lambda3 * curvTerm;
		}
		p.addP(crit,w,h); 
	    }
	    bsp.computeQ();
	    puissDeux *= 2; //
	}
    }
    /**
     * evolution criterion : log(1+sigmaIn^2) - log(1_sigmaOut^2) 
     *                     + ((I-muIn)^2-sigmaIn^2)/(1+sigmaIn^2) - ((I-muOut)^2-sigmaOut^2)/(1+sigmaOut^2) 
     *                     + curv
     * assuming that the object and the background have a constant variance
     **/
    public void evolveVariance1() {
	int s = listBsp.size();
	int [][] m;
	double muI, muO, sigmaI, sigmaO;

	m = createMask2();
	muO = meanOutside(m);
	sigmaO = varianceOutside(m,muO);
	System.out.println("Variance Exterieure = " + sigmaO);
	int puissDeux = 1; //

	for(int b = 0; b < s; b++) {
	    BSpline bsp = (BSpline) listBsp.get(b);
	    muI = meanInside(m,puissDeux); // muI = meanInside(m,b+1);
	    sigmaI = varianceInside(m,muI,puissDeux); //sigmaI = varianceInside(m,muI,b+1);
	    System.out.println("Variance Interieure = " + sigmaI);
	    int size = bsp.size();
	    for(int i = 0; i < size; i++) {
		BSPoint p = bsp.get(i);
		int I = getI((int) (p.getXP()),(int) (p.getYP()));
		double curvTerm = p.getCurv();
		if( curvTerm < 0) curvTerm = 0.0;
		double crit = lambda1 * ( Math.log(1+sigmaI) + ((I-muI)*(I-muI) -sigmaI)/(1 + sigmaI) ) - lambda2 * (Math.log(1+sigmaO) + ((I-muO)*(I-muO) -sigmaO)/(1+sigmaO)) + lambda3 * curvTerm;
		// 0.1 0.1 10
		p.addP(crit,w,h);
	    }
	    bsp.computeQ();
	    puissDeux *= 2; //
	}
    }
     /**
     * evolution criterion : 0 (no additive terms)
     *                     + ((I-muIn)^2-sigmaIn^2)/(1+sigmaIn^2) - ((I-muOut)^2-sigmaOut^2)/(1+sigmaOut^2) 
     *                     + curv
     * assuming that the object and the background have a constant variance
     **/
    public void evolveVariance2() {
	int s = listBsp.size();
	int [][] m;
	double muI, muO, sigmaI, sigmaO;

	m = createMask2();
	muO = meanOutside(m);
	sigmaO = varianceOutside(m,muO);
	int puissDeux = 1;

	for(int b = 0; b < s; b++) {
	    BSpline bsp = (BSpline) listBsp.get(b);
	    muI = meanInside(m,puissDeux);
	    sigmaI = varianceInside(m,muI,puissDeux);
	    int size = bsp.size();
	    for(int i = 0; i < size; i++) {
		BSPoint p = bsp.get(i);
		int I = getI((int) (p.getXP()),(int) (p.getYP()));
		double curvTerm = p.getCurv();
		if( curvTerm < 0) curvTerm = 0.0;
		double crit = lambda1 * ((I-muI)*(I-muI) -sigmaI)/(1 + sigmaI) - lambda2 *  ((I-muO)*(I-muO) -sigmaO)/(1+sigmaO) + lambda3 * curvTerm;
		p.addP(crit,w,h);
	    }
	    bsp.computeQ();
	    puissDeux *= 2; 
	}
    }
  
    /**
     * evolution criterion : lambda1 * (k * curv - grad(k).norm ) norm
     * with k = 1 / ( 1 + |grad(I)| )
     **/
    public void evolveGradient1() {
	int s = listBsp.size();

	for(int b = 0; b < s; b++) {
	    BSpline bsp = (BSpline) listBsp.get(b);
	    int size = bsp.size();
	    
	    for(int i = 0; i < size; i++) {
		BSPoint p = bsp.get(i);
		int xp = (int) (0.5 + p.getXP());
		int yp = (int) (0.5 + p.getYP());
		double Ix, Iy, Ixx, Iyy, Ixy;

		// gradient computation
		// Ix = (I(x+1)-I(x-1))/2   and Iy = (I(y+1)-I(y-1))/2
		// Ixx = (I(x+2) - 2*I(x)+ I(x-2))/4
		// Iyy = (I(y+2) - 2*I(y)+ I(y-2))/4
		// Ixy = (I(x+1,y+1)+I(x-1,y-1)-I(x+1,y-1)-I(x-1,y+1))/4

// 	if( (xp > 2)&&(xp < w-2)&&(yp > 2)&&(yp < h-2)) {
// 		    Ix = 0.5 * ( getI(xp+1, yp) - getI(xp-1, yp) );
// 		    Iy = 0.5 * ( getI(xp, yp+1) - getI(xp, yp-1) );
// 		    Ixx = 0.25 * (getI(xp-2, yp) + getI(xp+2, yp) -2*getI(xp, yp));
// 		    Iyy = 0.25 * (getI(xp, yp-2) + getI(xp, yp+2) -2*getI(xp, yp));
// 		    Ixy = 0.25 * (getI(xp+1, yp+1) + getI(xp-1, yp-1) - getI(xp-11, yp+1) - getI(xp+1, yp-1));
//              }
// 		else {Ix = 0; Iy = 0; Ixx = 0; Iyy = 0; Ixy = 0;}



// others computation method : fx(x) = f(x+1) - f(x), fxx(x) = fx(x+1)-fx(x) = f(x+2)-2f(x+1)+f(x)
//                             fxy(x) =   fx(x,y+1)-fx(x,y) = f(x+1,y+1)-f(x,y+1)-f(x+1,y)+f(x,y)
		if((xp < w-2)&&(yp < h-2)) {
		    Ix = getI(xp+1, yp) - getI(xp, yp);
		    Iy = getI(xp, yp+1) - getI(xp, yp);
		    Ixx = getI(xp+2, yp) - 2* getI(xp+1, yp) + getI(xp, yp);
		    Iyy = getI(xp, yp+2) - 2* getI(xp, yp+1) + getI(xp, yp);
		    Ixy = getI(xp+1, yp+1)  - getI(xp+1, yp) - getI(xp, yp+1) + getI(xp, yp);
		}
		else {
		    Ix = 0;
		    Iy = 0;
		    Ixx = 0;
		    Iyy = 0;
		    Ixy = 0;
		}
		double DI = Math.sqrt(Ix*Ix + Iy*Iy);
		double k = 1.0 / ( 1 + DI );

		// kx = -k^2/|gradI| (Ix*Ixx+Iy*Ixy)
		// ky = -k^2/|gradI| (Ix*Ixy+Iy*Iyy)

		double kx = 0.0, ky = 0.0, kN = 0.0;
		if( DI != 0) {
		    kx = - k*k*(Ix*Ixx+Iy*Ixy) / DI;
		    ky = - k*k*(Ix*Ixy+Iy*Iyy) / DI;
		    kN = p.getXNorm()*kx + p.getYNorm()*ky;
		}

		double crit = lambda1* (k * p.getCurv() - kN);
		if(Math.abs(crit) > Math.min(w,h)/3.0)   crit = 0;
		//System.out.println("crit = " + crit);
		p.addP(crit,w,h);
	    }
	    bsp.computeQ();
	}
    }
   /**
     * evolution criterion : lambda1 * (k * curv - grad(k).norm ) norm
     * with k = 1 / ( 1 + |grad(I)|^2 )
     **/
    public void evolveGradient2() {
	int s = listBsp.size();

	for(int b = 0; b < s; b++) {
	    BSpline bsp = (BSpline) listBsp.get(b);
	    int size = bsp.size();
	    
	    for(int i = 0; i < size; i++) {
		BSPoint p = bsp.get(i);
		int xp = (int) (0.5 + p.getXP());
		int yp = (int) (0.5 + p.getYP());
		double Ix, Iy, Ixx, Iyy, Ixy;

		if((xp < w-2)&&(yp < h-2)) {
		    Ix = getI(xp+1, yp) - getI(xp, yp);
		    Iy = getI(xp, yp+1) - getI(xp, yp);
		    Ixx = getI(xp+2, yp) - 2* getI(xp+1, yp) + getI(xp, yp);
		    Iyy = getI(xp, yp+2) - 2* getI(xp, yp+1) + getI(xp, yp);
		    Ixy = getI(xp+1, yp+1)  - getI(xp+1, yp) - getI(xp, yp+1) + getI(xp, yp);
		}
		else {
		    Ix = 0;    Iy = 0;    Ixx = 0;    Iyy = 0;    Ixy = 0;
		}
		double DI2 = Ix*Ix + Iy*Iy;
		double k = 1.0 / ( 1 + DI2 );

		double kx = 0.0, ky = 0.0, kN = 0.0;
		if( DI2 != 0) {
		    kx = - 2.0*k*k*(Ix*Ixx+Iy*Ixy);
		    ky = - 2.0*k*k*(Ix*Ixy+Iy*Iyy);
		    kN = p.getXNorm()*kx + p.getYNorm()*ky;
		}

		double crit = lambda1* (k * p.getCurv() - kN);
		if(Math.abs(crit) > Math.min(w,h)/3.0)   crit = 0;
		p.addP(crit,w,h);
	    }
	    bsp.computeQ();
	}
    }
   /**
     * evolution criterion : lambda1 * k * (lambda2 + curv) norm
     * with k = 1 / ( 1 + |grad(I)| )
     **/
    public void evolveGradient3() {
	int s = listBsp.size();

	for(int b = 0; b < s; b++) {
	    BSpline bsp = (BSpline) listBsp.get(b);
	    int size = bsp.size();
	    
	    for(int i = 0; i < size; i++) {
		BSPoint p = bsp.get(i);
		int xp = (int) (0.5 + p.getXP());
		int yp = (int) (0.5 + p.getYP());
		double Ix, Iy;

		if((xp < w-1)&&(yp < h-1)) {
		    Ix = getI(xp+1, yp) - getI(xp, yp);
		    Iy = getI(xp, yp+1) - getI(xp, yp);
		}
		else {		    Ix = 0;    Iy = 0;		}
		double DI = Math.sqrt(Ix*Ix + Iy*Iy);
		double k = 1.0 / ( 1 + DI );
		double crit = lambda1* k * (p.getCurv() + lambda2);
		if(Math.abs(crit) > Math.min(w,h)/3.0)   crit = 0;
		p.addP(crit,w,h);
	    }
	    bsp.computeQ();
	}
    }

   /**
     * evolution criterion : lambda1 * k * (lambda2 + curv) norm
     * with k = 1 / ( 1 + |grad(I)|^2 )
     **/
    public void evolveGradient4() {
	int s = listBsp.size();

	for(int b = 0; b < s; b++) {
	    BSpline bsp = (BSpline) listBsp.get(b);
	    int size = bsp.size();
	    
	    for(int i = 0; i < size; i++) {
		BSPoint p = bsp.get(i);
		int xp = (int) (0.5 + p.getXP());
		int yp = (int) (0.5 + p.getYP());
		double Ix, Iy;

		if((xp < w-1)&&(yp < h-1)) {
		    Ix = getI(xp+1, yp) - getI(xp, yp);
		    Iy = getI(xp, yp+1) - getI(xp, yp);
		}
		else {		    Ix = 0;    Iy = 0;		}
		double DI2 = Ix*Ix + Iy*Iy;
		double k = 1.0 / ( 1 + DI2 );
		double crit = lambda1* k * (p.getCurv() + lambda2);
		if(Math.abs(crit) > Math.min(w,h)/3.0)   crit = 0;
		p.addP(crit,w,h);
	    }
	    bsp.computeQ();
	}
    }

    /**
    * computes and returns the mean of the images points that are inside the BSpline of color color, 
    * using the mask m given as a parameter.
    * @param m mask already computed
    * @param color the color used for this region
    **/
    public double meanInside(int [][] m, int color) {
	double mean = 0;
	int nbPoints = 0;

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) 
		if((m[i][j] & color) == color) { // we are inside the region of color color
		    mean += getI(i,j);
		    nbPoints ++;
		}
	if(nbPoints == 0) return -1;
	mean /= nbPoints;
	System.out.println(" mean = " + mean + " with " + nbPoints + " points.");
	return mean;
    }
    /**
    * computes and returns the variance of the images points that are inside the BSpline of color color, using the mask m
    * given as a parameter and the mean mu.
    **/
     public double varianceInside(int [][] m, double mu, int color) {
	double sigma = 0;
	int nbPoints = 0;

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) 
		if(m[i][j] == color) { // we are inside the region of color color
		    int col = getI(i,j);
		    sigma += (col - mu) * (col - mu);
		    nbPoints ++;
		}
	if(nbPoints == 0) return -1;
	sigma /= nbPoints;
	return sigma;
    }
    /**
    * computes and returns the mean of the images points that are outside the BSpline, using the mask m
    * given as a parameter.
    **/
    public double meanOutside(int [][] m) {
	double mean = 0;
	int nbPoints = 0;

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) 
		if(m[i][j] == 0) { // we are outside
		    mean += getI(i,j);
		    nbPoints ++;
		}
	if(nbPoints == 0) return -1;
	mean /= nbPoints;
	return mean;
    }
   /**
    * computes and returns the variance of the images points that are outside the BSpline, using the mask m
    * given as a parameter and the mean mu.
    **/
     public double varianceOutside(int [][] m, double mu) {
	double sigma = 0;
	int nbPoints = 0;

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) 
		if(m[i][j] == 0) { // we are outside
		    int color = (getI(i,j));
		    sigma += (color - mu) * (color - mu);
		    nbPoints ++;
		}
	if(nbPoints == 0) return -1;
	sigma /= nbPoints;
	return sigma;
    }

    /**
     * checks if two BSplines are overlapping or not.
     * returns 0 if they are not overlapping (80 %)
     * returns 1 if the 1st BSplines contains approximately the second
     * returns 2 if the 2nd BSplines contains approximately the first
     * 
     **/
    public int areOverlapping(int [][] m, int color1, int color2) {
	int only1 = 0;
	int only2 = 0;
	int oneAndTwo = 0;
	
	for(int i = 0; i < m.length; i++) {
	    for(int j = 0; j < m[0].length; j++) {
		if( ( (m[i][j] & color1)== color1) && ( (m[i][j] & color2)== color2))
		    oneAndTwo ++;
		else if((m[i][j] & color1)== color1) 
		    only1 ++;
		else if((m[i][j] & color2)== color2) 
		    only2 ++;
	    }
	}
	if(oneAndTwo == 0)
	    return 0;
	if( 100.0/DIFF_RATIO*Math.abs(only2-only1)/(Math.max(only1,only2)+oneAndTwo) < 1) 
	    if(only1 > only2)
		return 1;
	    else
		return 2;
	return 0;
    }

    /**
     * examines all the Splines and eliminates those that are overlapping >- 80% another
     **/
    public void checkOverlapping() {
	int [][] m;
	m = createMask2(); 
	int s = listBsp.size();

	if(s < 2)
	    return;

	int [] puissDeux = new int[s];
	boolean [] toKeepTab = new boolean[s];
	
	puissDeux[0] = 1;
	toKeepTab[0] = true;
	
	for(int i = 1; i < s; i++) {
	    puissDeux[i] = puissDeux[i-1]*2;
	    toKeepTab[i] = true;
	}
	
	for(int i = 0; i < s - 1; i++) {
	    if(toKeepTab[i]) {
		for(int j = i+1; j < s; j++) {
		    if(toKeepTab[j]) {
			int toKeep = areOverlapping(m, puissDeux[i], puissDeux[j]);
			if(toKeep == 1)
			    toKeepTab[j] = false;
			else if(toKeep == 2) {
			    toKeepTab[i] = false;
			    j = s;
			}
		    }
		}
	    }
	}
	for(int i = s-1 ; i >= 0; i--)
	    if(!toKeepTab[i])
		listBsp.remove(i);
    }
	
    public int [][] createMask2() {
	int s = listBsp.size();
	int [][] res = new int[w][h];
	int puissDeux = 1;
	
	if( s == 0) return res;

	for(int i = 0; i < s; i++) {
	    BSpline bsp = (BSpline) listBsp.get(i);
	    bsp.createMask(res, 0, 0, w, h, puissDeux);
	    puissDeux *= 2;
	}
	return res;
    }

    /**
    * paints this graphical component and the BSpline corresponding
    **/
    public void paintComponent(Graphics gc) {
	gc.setColor(Color.WHITE);
	super.paintComponent(gc);
	Graphics2D g2 = (Graphics2D) gc;
	if(bi != null) {
	    g2.drawImage(bi,null,0,0);
	    
	    if(withMask) {
		int [][] m = createMask2();
		for(int i = 0; i < w; i++)
		    for(int j = 0; j < h; j++)
			if(m[i][j] > 0) {
			    gc.setColor(BS_COLORS[(BS_COLORS.length + m[i][j]-1)%BS_COLORS.length]);
			    gc.fillRect(i,j,1,1);
			}
			else if (m[i][j] < 0) { // ceci n'arrive que lorsqu'on a + de 30 regions
			    // et que 2^31 est negatif . De toutes facons, on perd toutes les autres
			    // regions
			    gc.setColor(BS_COLORS[0]);
			    gc.fillRect(i,j,1,1);
			}
	    }
	}
	int s = listBsp.size();
	System.out.println("drawing " + s + " bsplines");
	for(int i = 0; i < s ; i++ ) {
	    BSpline bsp = (BSpline) listBsp.get(i);
	    //bsp.paint(gc,Color.WHITE);
	    bsp.paint(gc,BS_COLORS[(BS_COLORS.length + i)%BS_COLORS.length]);
	}
    }

    public void distMapShow() {}
}
