package bsplines;

/*
Tutorial on LevelSet and BSpline deformable model techniques 
for image segmentation
Copyright (C) Diane Lingrand (Diane.Lingrand@unice.fr)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

import java.awt.Graphics;
import java.awt.Color;

/**
 * class for BSPline elements that are composed from :
 * - a P-point (point on the curve)
 * - a Q-point (control point)
 * - a segment of curve from the P-point of this segment to the P-point of the next segment
 * The polynomial coefficients of the curve segment are computed from several Q-points so that we
 * cannot compute them in this class. These are the instance variables a[] and b[].
 * We need also the normal vector and the curvature at the P-point.
 *
 * @author <a href="mailto:Diane.Lingrand@unice.fr">DianeLingrand</a>
 **/
public class BSPoint {
    
    private static final int CROSS = 5; // radius of cross drawn at P position
    private static final int LINE_WIDTH=4; // width of line when drawing on screen
    private static final int LINE_OFFSET=2; // half width of line when drawing on screen
   private double xP;   // point P coordinates
    private double yP;
    private double xQ;   // point Q coordinates (control point)
    private double yQ;
    private double xNorm;  // normal at P point
    private double yNorm;
    private double curv;   // curvature at P point
    private double [] a = new double[4]; // x coefficients
    private double [] b = new double[4]; // y coefficients

    /**
    * constructor : with all instance variables 
    **/
    public BSPoint(double _xP, double _yP, double _xQ, double _yQ, double _xNorm, double _yNorm, double _curv, double [] _a, double [] _b) {
	xP = _xP;
	yP = _yP;
	xQ = _xQ;
	yQ = _yQ;
	xNorm = _xNorm;
	yNorm = _yNorm;
	curv = _curv;
	for(int i = 0; i < 4 ; i++) {
	    a[i] = _a[i];
	    b[i] = _b[i];
	}
    }

    /**
    * constructor : of a P-point (we will compute the other variable further)
    **/
    public BSPoint(double _xP, double _yP) {
	xP = _xP;
	yP = _yP;
	a[0] = xP;
	b[0] = yP;
    }

    /**
    * text formatting of this BSPoint
    **/
    public String toString() {
	String result="";
	
	result += " Point P = ( " + xP + " ; " + yP + " )\n" ;
	result += "    with Point Q = ( " + xQ + " ; " + yQ + " ), " ;
	result += "normal = ( " + xNorm + " ; " + yNorm + " ), " ;
	result += "curvature = " + curv + ", ";
	result += "a0 = " + a[0] + " ;  b0 = " + b[0];

	return result;
    }

    /// accessors
    public double getXP() {	return xP; }
    public double getYP() {	return yP; }
    public double getCurv() {   return curv; }
    public double getXNorm() {  return xNorm;}
    public double getYNorm() {  return yNorm;}

    /**
    * returns the P point corresponding to t = 0.5
    * used in method void checkDistances() from class Bsplines
    **/
    public BSPoint getMiP() {
	double t = 0.5;
	double x = a[0] + t * ( a[1] + t * ( a[2] + t * a[3]));
	double y = b[0] + t * ( b[1] + t * ( b[2] + t * b[3]));
	return new BSPoint(x,y);
    }

    /// modifiers
    public void setP( double newX, double newY) {	xP = newX;	yP = newY;    a[0] = xP; b[0] = yP; }
    /**
     * adds to this P-point the quantities newX to xP and newY to yP.
     * This translates P by [newX newY].
     **/
    public void addP( double newX, double newY) {	
	xP += newX;	yP += newY;   
	a[0] = xP;                          b[0] = yP; 
    }
    /**
     * adds to this P-point the quantities newX to xP and newY to yP.
     * This translates P by [newX newY].
     * Verifies that the resulting point is still in the window (xmax x yMax).
     **/
    public void addP( double newX, double newY, int xMax, int yMax) {	
	xP += newX;	
	xP = (xP < xMax) ? (xP < 0) ? 0 : xP : xMax-1;
	yP += newY;   
	yP = (yP < yMax) ? (yP < 0) ? 0 : yP : yMax-1;
	a[0] = xP; b[0] = yP;
    }
    /**
     * adds to this P-point a quantity "a" multiplied by the normal vector at P
     **/
    public void addP( double a) {	
	xP += a*xNorm;	yP += a*yNorm;    
    }
    /**
     * adds to this P-point a quantity "a" multiplied by the normal vector at P.
     * Verifies that the resulting point is still in the window (xmax x yMax).
     **/
    public void addP( double a, int xMax, int yMax) {	
	xP += a*xNorm;	
	xP = (xP < xMax) ? (xP < 0) ? 0 : xP : xMax-1;
	yP += a*yNorm;    
	yP = (yP < yMax) ? (yP < 0) ? 0 : yP : yMax-1;
	this.a[0] = xP; this.b[0] = yP;
    }

    public void setQ( double newX, double newY) {	xQ = newX;	yQ = newY;    }
    public void setNorm( double newX, double newY) {    xNorm = newX;   yNorm = newY; }
    public void setCurv( double newC) {	curv = newC;   }
    /**
     * affects the a[] and b[] instances variables
     * and (re)computes the normal and curvature from these new values
     **/ 
    public void setAB( double [] newA, double [] newB) {
	for(int i = 0 ; i < 4 ; i ++) {
	    a[i] = newA[i];
	    b[i] = newB[i];
	}
	computeNorm();
	computeCurv();
    }
    
    /**
    * computes and sets the normal vector at the P-point
    **/
     public void computeNorm() {
	double sqrt = Math.sqrt(a[1]*a[1] + b[1]*b[1]);
	xNorm = -b[1] / sqrt;
	yNorm = a[1] / sqrt;
    }
    
    /**
    * computes and sets the curvature at the P-point
    **/
    public void computeCurv() {
	double sqr = a[1]*a[1] + b[1]*b[1];
	curv = 2.0 * (a[1]*b[2] - a[2]*b[1]) / Math.sqrt(sqr*sqr*sqr);
    }

    /**
    * returns the euclidean distance between this P-point and the P-point of argument
    * 
    * @param p distant point
    **/
    public double dist(BSPoint p){
	double x = xP - p.getXP();
	double y = yP - p.getYP();
	return Math.sqrt(x*x + y*y);
    }

    /**
    * returns the approximation of arc length between t=0 and t=1 using recursive dichotomie
    * with "precision" epsilon
    * 
    * @param epsilon related to the precision of the evaluated length of the curve
    **/
    public double length(double epsilon) {
	return length(0.0, 1.0, epsilon);
    }
    /**
    * returns the approximation of arc length between t0 and t1 using recursive dichotomie
    * with "precision" epsilon
    * This is done assuming that 0 <= t0 <= t1 <= 1
    * 
    * @param epsilon related to the precision of the evaluated length of the curve
    * @param t0  
    * @param t1
    **/
    public double length(double t0, double t1, double epsilon) {
	//trivial case
	if(t0 == t1)
	    return 0;
	
	double tm = (t0 + t1)*0.5;
	BSPoint p0 = pointAtTvalue(t0);
	BSPoint p1 = pointAtTvalue(t1);
	BSPoint pm = pointAtTvalue(tm);
	double d01 = p0.dist(p1);
	double d0m = p0.dist(pm);
	double dm1 = pm.dist(p1);
	if(Math.abs(d01 - (d0m+dm1)) < epsilon)
	    return d0m+dm1;
	return (length(t0,tm,epsilon) + length(tm,t1,epsilon));
    }

    /**
     * finds and returns the t value corresponding to the P point x(t),y(t)
     * verifying that |dist (P0, P) - value | < epsilon
     **/
    public double findT(double debut, double value, double epsilon) {
	double tRes = debut;
	if((debut<0.0)||(debut>1.0))
			 return -1.0;
	BSPoint init = this.pointAtTvalue(debut);
	tRes = dichotomie(init, debut, 1.0, value, epsilon);
	return tRes;
    }

    // used in method findT(..,)
    private double dichotomie(BSPoint binit, double debut, double fin, double value, double epsilon) {
	double milieu = (debut + fin) * 0.5;

	if(fin-milieu == 0) return milieu;
	if(milieu-debut == 0) return milieu;
	
	double vmilieu = binit.dist(this.pointAtTvalue(milieu));
	if(vmilieu < value - epsilon)
	    return dichotomie(binit, milieu, fin, value, epsilon);
	else if(vmilieu > value + epsilon)
	    return dichotomie(binit, debut, milieu, value, epsilon);
	else 
	    return milieu;
    }

    /**
     * computes and returns the P-point located at value t on this BSpline segment
     **/
    public BSPoint pointAtTvalue(double t) {
	double x = a[0] + t * ( a[1] + t * ( a[2] + t * a[3]));
	double y = b[0] + t * ( b[1] + t * ( b[2] + t * b[3]));
	return new BSPoint(x,y);
    }

    /** 
     * clones this BSPoint
     **/
    public java.lang.Object clone() {
	return new BSPoint(xP, yP, xQ, yQ, xNorm, yNorm, curv, a, b);
    }

    /**
     * draws the curve on a 2D array (used for the mask creation in BSpline) using the color color, with thickness 2.
     **/
    public void trace(int [][] m, int color) {
	double d = Math.sqrt((a[1]+a[2]+a[3])*(a[1]+a[2]+a[3]) + (b[1]+b[2]+b[3])*(b[1]+b[2]+b[3]));
	double drawStep = 0.5/d;

	for(double t = 0.0; t < 1; t += drawStep) {
	    int x = (int) (a[0] + t * ( a[1] + t * ( a[2] + t * a[3])));
	    int y = (int) (b[0] + t * ( b[1] + t * ( b[2] + t * b[3])));
	    x = (x<m.length-1) ? x : (m.length-2) ;
	    x = (x<0) ? 0 : x;
	    y = (y<0) ? 0 : (y<m[0].length-1) ? y : (m[0].length-2) ;
	    y = (y<0) ? 0 : y;
	    m[x][y] = color;
	    m[x+1][y] = color;
	    m[x][y+1] = color;
	    m[x+1][y+1] = color;

	}
    }

    /**
     * draws the curve on a 2D array (used for the self intersection detection) using the color color.
     *
     * @param x0 y0 coordinates of the first point P0 of the BSPline
     * @return value used for intersection detection
     **/
    public int traceB(int [][] m, int color, int x0, int y0) {
	int res = -1;
	int xm1,ym1,xp1,yp1;
	int x, y, xFirst, yFirst, xOld, yOld;
	int w = m.length;
	int h = m[0].length;

	double d = Math.max(Math.abs(a[1]+a[2]+a[3]),Math.abs(b[1]+b[2]+b[3]));
	double drawStep = 0.5/d;

	// first drawn point

	xFirst = (int) (0.5 + a[0]); 
	yFirst = (int) (0.5 + b[0]); 
	xFirst = (xFirst < w) ? ( (xFirst<0)? 0 : xFirst) : (w-1) ;
	yFirst = (yFirst < h) ? ( (yFirst<0)? 0 : yFirst) : (h-1) ;

	x = xFirst;
	y = yFirst;

	if(color == 1) { // we paint the first point of the first segment
	    if((m[x][y]!= 0) && (m[x][y]!=color))
		return m[x][y];
	}
	else { // we paint the first point of at least the second segment
	    if((m[x][y]!= 0) && (m[x][y]!=color) && (m[x][y]!=color-1))
		return m[x][y];
	}
	m[x][y] = color;
	double t = 0.0;
	while( (x == xFirst) && (y == yFirst) && (t < 1-drawStep)) {
	    t += drawStep;
	    x = (int) (0.5 + a[0] + t * ( a[1] + t * ( a[2] + t * a[3])));
	    y = (int) (0.5 + b[0] + t * ( b[1] + t * ( b[2] + t * b[3])));
	    x = (x<w) ? ((x<0) ? 0 : x) : (w-1) ;
	    y = (y<h) ? ((y<0) ? 0 : y) : (h-1) ;
	}
	if( (x == xFirst) && (y == yFirst) )
	    return res;
	xOld = xFirst;
	yOld = yFirst;

	// (x,y) is the next point to examine, different from the first point

	while ( t < 1 ) {
	    // if the current point is the first point, P0 : exit
	    if( (x == x0) && (y == y0) )
		return res;

	    // drawing the current point
	    if((m[x][y] != 0) && (m[x][y] != color))
		return m[x][y];

	    // looking of X intersections between segment as for example
	    //       1 0 0 2 0
	    //       0 1 2 0 0
	    //       0 2 1 0 0
	    //       2 0 0 1 0
	    if( (x != xOld) && (y != yOld) ) { // not case | or _
		if((m[x][yOld] != 0) && (m[x][yOld] != color))
		    return m[x][yOld];
		if((m[xOld][y] != 0) && (m[xOld][y] != color))
		    return m[xOld][y];
	    }
	    m[x][y] = color;
	    
	    // computing the next different point
	    xOld = x;
	    yOld = y;

	    while( (x == xOld) && ( y == yOld) && (t < 1 - drawStep) ) {
		t += drawStep;
		x = (int) (0.5 + a[0] + t * ( a[1] + t * ( a[2] + t * a[3])));
		y = (int) (0.5 + b[0] + t * ( b[1] + t * ( b[2] + t * b[3])));
		x = (x<m.length-1) ?    ((x<0) ? 0 : x) : (m.length-2) ;
		y = (y<m[0].length-1) ? ((y<0) ? 0 : y) : (m[0].length-2) ;
	    }
	    if((x == xOld) && ( y == yOld)) // no next point on this segment
		return res;
	}
	return res;
    }

    // the same with a double [][], thickness 1 and no intersection searching
    public void traceB(double [][] m, int color) {
	int x, y;
	
	double d = Math.max(Math.abs(a[1]+a[2]+a[3]), Math.abs(b[1]+b[2]+b[3]));
	double drawStep = 0.5/d;

	for(double t = 0; t <= 1; t += drawStep) {
	    x = (int) (0.5 + a[0] + t * ( a[1] + t * ( a[2] + t * a[3])));
	    y = (int) (0.5 + b[0] + t * ( b[1] + t * ( b[2] + t * b[3])));
	    x = (x<m.length-1) ?    ((x<0) ? 0 : x) : (m.length-2) ;
	    y = (y<m[0].length-1) ? ((y<0) ? 0 : y) : (m[0].length-2) ;
	  
	    m[x][y] = color;
	    
	}
    }

    /**
    * draws the curve in a Graphics g
    **/
    public void paint(Graphics g, Color c) {
	// draw the curve from P to P+1

	// d is the euclidean distance from P to P+1
	double d = Math.sqrt((a[1]+a[2]+a[3])*(a[1]+a[2]+a[3]) + (b[1]+b[2]+b[3])*(b[1]+b[2]+b[3]));
	// we will trace E(2*d) points between P and P+1
	double drawStep = 0.5 / d;

	g.setColor(c);

	for(double t = 0.0; t < 1; t += drawStep) {
	    double x = a[0] + t * ( a[1] + t * ( a[2] + t * a[3]));
	    double y = b[0] + t * ( b[1] + t * ( b[2] + t * b[3]));
	    g.fillOval((int) (x+0.5)-LINE_OFFSET, (int) (y+0.5)-LINE_OFFSET, LINE_WIDTH, LINE_WIDTH);
	    if( (int) (x+0.5) > 0 ) {
		g.fillOval((int) (x-0.5)-LINE_OFFSET, (int) (y+0.5)-LINE_OFFSET, LINE_WIDTH, LINE_WIDTH);
		if( (int) (y+0.5) > 0 )
		    g.fillOval((int) (x-0.5)-LINE_OFFSET, (int) (y-0.5)-LINE_OFFSET, LINE_WIDTH, LINE_WIDTH);
	    }
	    else
		if( (int) (y+0.5) > 0 )
		    g.fillOval((int) (x+0.5)-LINE_OFFSET, (int) (y-0.5)-LINE_OFFSET, LINE_WIDTH, LINE_WIDTH);
	}

	// draw a cross at P position
	g.setColor(Color.RED);
	// TODO et pourquoi pas un cercle ?
	//	g.drawLine((int) xP - CROSS, (int) yP, (int) xP + CROSS, (int) yP);
	//	g.drawLine((int) xP, (int) yP - CROSS, (int) xP, (int) yP + CROSS);
	g.fillOval((int) xP - CROSS,(int) yP - CROSS,2*CROSS,2*CROSS);

	// draw the normal (with norm = 20)
	//g.setColor(Color.GREEN);
	g.setColor(Color.RED);

	for(int i=0; i<20;i++) {
	    g.fillOval((int) (xP+i*xNorm), (int) (yP+i*yNorm),2,2);
	}
	//g.drawLine((int) xP, (int) yP, (int) (xP+20*xNorm), (int) (yP+20*yNorm));

	// draw the control points
	//	g.setColor(Color.PINK);
	//	g.drawLine((int) xQ - CROSS, (int) yQ, (int) xQ + CROSS, (int) yQ);
	//	g.drawLine((int) xQ, (int) yQ - CROSS, (int) xQ, (int) yQ + CROSS);
	
    }
    

}
