package bsplines;

/*
Tutorial on LevelSet and BSpline deformable model techniques 
for image segmentation
Copyright (C) Diane Lingrand (Diane.Lingrand@unice.fr)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

import java.util.ArrayList;
import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.image.*;
import javax.imageio.*;
import java.io.*;
import java.util.jar.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.awt.Point;

/**
 * LSPanel class :
 *    manages the evolution of a Level-Set according to a BufferedImage
 * 
 * @author <a href="mailto:Diane.Lingrand@unice.fr">DianeLingrand</a>
 **/

public class LSPanel extends SegPanel implements MouseListener{
   
    private double [][] levelSet;
    private boolean inConstruction = true;
    private boolean withMask = false; // do we draw the mask ?
    private boolean distMapShow = false;

    private int iter = 0; // using for name of saved image in regulDistances()
 
    public LSPanel(double l1, double l2, double l3) {
	this(IMAGE_FILENAME, l1, l2, l3);
    }

    public LSPanel(String filename, double l1, double l2, double l3) {
	super();

	lambda1 = l1;
	lambda2 = l2;
	lambda3 = l3;

	this.addMouseListener(this);
	setPreferredSize(new Dimension(w,h));
	setBackground(new Color(255,255,255));

	// opening an image in this panel
	URL imageURL = bsplines.BSApplet.class.getResource("images/"+IMAGE_FILENAME);
	System.out.println("url = " + imageURL);
	if(imageURL != null)
	    changeImage(imageURL);
	if(bi != null) {
	    w = bi.getWidth();
	    h = bi.getHeight();
	    biLiss = lissage(bi);
	    setPreferredSize(new Dimension(w,h));
	    System.out.println("image read with w = " + w + " and h = " + h);
	    inConstruction = false;
	}
	else
	    System.out.println("Unable to open file " + filename);

	levelSet = new double[w][h];
	//init circle centered in w/2, h/2, radius min(w,h)/2.5
	double radius = Math.min(w,h)/2.5;
	double xCenter = w/2;
	double yCenter = h/2;
	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) {
		levelSet[i][j] = Math.sqrt( (i - xCenter)*(i - xCenter) + (j - yCenter)*(j - yCenter)) - radius;
	    }
    }

    public LSPanel(BSPanel pan) {
	super();

	lambda1 = pan.getLambda1();
	lambda2 = pan.getLambda2();
	lambda3 = pan.getLambda3();
	bi = pan.getBufferedImage();
	biLiss = pan.getLissBufferedImage();
	avecLissage = pan.getSmoothing();
	capture = pan.getCapture();

	ArrayList list = pan.getListBsp();

	this.addMouseListener(this);
	setPreferredSize(new Dimension(220,220));
	setBackground(new Color(255,255,255));

	w = bi.getWidth();
	h = bi.getHeight();
	setPreferredSize(new Dimension(w,h));
	
	levelSet = new double[w][h];
	constructLevelSet(list);
    }

    /**
     * used in the levelset construction from bsplines:
     * computes the distance map from the drawing of the splines
     **/
    private void constructLevelSet(ArrayList list) {
	int max = Math.max(w,h);
	int s = list.size();
	int wh = w*h;
	System.out.println(" wh = " + wh);
	// init levelSet to the value + max
	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++)
		levelSet[i][j] = wh;

	// then we draw the BSplines with color = -1
	for(int i = 0; i < s; i++) {
	    BSpline bsp = (BSpline) list.get(i);
	    int size = bsp.size();
	    for(int j = 0; j < size; j++) {
		BSPoint p = (BSPoint) bsp.get(j);
		p.traceB(levelSet,-1);
	    }
	}

	// we need now to mark the outside with + max and inside - max
	// to do that, we will color with positive colors the differents regions
	// the first region is outside the forms with color 0
	int color = 0;
	int [] corresp = new int[wh];
	for(int i = 0; i < wh; i++) 
	    corresp[i] = i;

	// we assume that the point [0][0] is outside the spline. then, the
	// color of this point is 0
	levelSet[0][0] = color; // = 0
	
	//// FIRST PASS
	// case of the first line.
	for(int j = 1; j < h; j++)
	    if(levelSet[0][j] >= 0){
		if(levelSet[0][j-1] >= 0)
		    levelSet[0][j] = levelSet[0][j-1];
		else
		    levelSet[0][j] = ++color;
	    }

	// other lines i
	for(int i = 1; i < w; i++) {
	    // case of j=0
	    if(levelSet[i][0] >= 0) {
		if(levelSet[i-1][0] >= 0)
		    levelSet[i][0] = levelSet[i-1][0];
		else
		    levelSet[i][0] = ++color;
	    }
	    // other columns j
	    for(int j = 1; j < h; j++) {
		if(levelSet[i][j] >= 0) {
		    if(levelSet[i-1][j] >= 0) {
			levelSet[i][j] = levelSet[i-1][j];
			if( (levelSet[i][j-1] >= 0) && (levelSet[i][j] != levelSet[i][j-1]) ) {
			    // two different colors correspond to the same region
			    int mi;
			    int ma;
			    if(levelSet[i][j-1] < levelSet[i][j]) {
				mi = (int) levelSet[i][j-1];
				ma = (int) levelSet[i][j];
			    }
			    else {
				mi = (int) levelSet[i][j];
				ma = (int) levelSet[i][j-1];
			    }
			    levelSet[i][j] = mi; 
			    tryCorresp(corresp,mi,ma); //corresp[ma] = mi;
			}
		    }
		    else if(levelSet[i][j-1] >= 0)
			levelSet[i][j] = levelSet[i][j-1];
		    else
			levelSet[i][j] = ++color;
		}
	    }
	}
	//// UPDATING the table of color correspondances
	boolean updated = false;
	while( !updated ) {
	    updated = true;
	    for(int i = 0; i < wh; i++) 
		if(corresp[corresp[i]] != corresp[i]) {
		    corresp[i] = corresp[corresp[i]];
		    updated = false;
		}
	}

	//// SECOND PASS
	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) {
		// use the table of color correspondances
		if(levelSet[i][j] >= 0)
		    levelSet[i][j] = corresp[(int) levelSet[i][j]];
		// +max for outside(0), -max for inside (>0), 0 for border (-1)
		if(levelSet[i][j] > 0)
		    levelSet[i][j] = - max;
		else if(levelSet[i][j] == 0)
		    levelSet[i][j] = + max;
		else if(levelSet[i][j] == -1)
		    levelSet[i][j] = 0;

	    }

	// finally, we compute the distance map
	this.regulDistances(0);

   }

    public double [][] getLevelSet() { return levelSet;}

    /**
     * save the parameters in an ASCII file such as :
     * close distant
     * criterion
     * lambda1 lambda2 lambda3
     **/
    public void saveParameters(File file) {
	try{
	    BufferedWriter writer = new BufferedWriter(new FileWriter(file));

	    writer.write(""+ criterion + "\n");
	    writer.write(""+ lambda1 + " " + lambda2 + " " + lambda3 + "\n");
	    writer.flush();
	    writer.close();
	}
	catch(IOException e) {
	    System.out.println("Unable to write in file " + file);
	}
    }
     public void saveParameters(JTextArea jt) {
	    jt.append(""+ criterion + "\n");
	    jt.append(""+ lambda1 + " " + lambda2 + " " + lambda3 + "\n");
    }

    
    /** 
     * clears all the LevelSets in this Panel
     **/
    public void clearBSplines() {
	double radius = Math.min(w,h)/2.5;
	double xCenter = w/2;
	double yCenter = h/2;
	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) {
		levelSet[i][j] = Math.sqrt( (i - xCenter)*(i - xCenter) + (j - yCenter)*(j - yCenter)) - radius;
	    }
	repaint();	       
    }
    public void changeDrawWithMask() {
	withMask = !withMask;
	repaint();
    }

    public void distMapShow() {
	distMapShow = ! distMapShow;
	repaint();
    }

    // implements ...
    public void mouseClicked(MouseEvent e){}
    public void mouseEntered(MouseEvent e){}
    public void mouseExited(MouseEvent e){}

    public void mousePressed(MouseEvent e){
	if(e.getButton() == MouseEvent.BUTTON2)
	    changeDrawWithMask();
	else if(e.getButton() == MouseEvent.BUTTON1)
	    evolve(true);
    }
    
    public void mouseReleased(MouseEvent e){
	if( e.getButton() == MouseEvent.BUTTON1)
	    evolve(false);
    }

    /**
     * evoluting the BSpline according to image informations.
     **/
    public void evolve(boolean b) {

	if(b && (bi == null)) {
	    paintImmediately(0, 0, (int) getSize().getWidth(), (int) getSize().getHeight());
	    return;
	}

	// mouse down
	if(b)  applyCrit();
	// mouse up
	else   regulDistances(0);

	paintImmediately(0, 0, (int) getSize().getWidth(), (int) getSize().getHeight());
    }

    /**
     * evolution criterion : (I-muIn)^2 - (I-muOut)^2 + curv
     * assuming that the object and the background are homogeneous
     **/
    public void evolveMean() {
	double muI, muO;
	double crit;

	muO = meanOutside();
	muI = meanInside();

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) {
		int I = getI(i,j);
		double curvTerm = getCurv(i,j);
		if( curvTerm < 0) curvTerm = 0.0;
		crit = lambda1 * (I-muI)*(I-muI) - lambda2 * (I-muO)*(I-muO) + lambda3 * curvTerm;
		if(Math.abs(crit) > Math.max(w,h)/2)
		    crit = Math.max(w,h)/2.0;
		levelSet[i][j] += crit;
	    }
    }
    /**
     * evolution criterion : log(1+sigmaIn^2) - log(1_sigmaOut^2) 
     *                     + ((I-muIn)^2-sigmaIn^2)/(1+sigmaIn^2) - ((I-muOut)^2-sigmaOut^2)/(1+sigmaOut^2) 
     *                     + curv
     * assuming that the object and the background have a constant variance
     **/
    public void evolveVariance1() {
	double muI, muO, sigmaI, sigmaO;

	muO = meanOutside();
	muI = meanInside();
	sigmaO = varianceOutside(muO);
	sigmaI = varianceInside(muI);

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) {
		int I = getI(i,j);
		double curvTerm = getCurv(i,j);
		if( curvTerm < 0) curvTerm = 0.0;
		double crit = lambda1 * ( Math.log(1+sigmaI) + ((I-muI)*(I-muI) -sigmaI)/(1 + sigmaI) ) - lambda2 * (Math.log(1+sigmaO) + ((I-muO)*(I-muO) -sigmaO)/(1+sigmaO)) + lambda3 * curvTerm;
		levelSet[i][j] += crit;
	    }

    }
     /**
     * evolution criterion : 0 (no additive terms)
     *                     + ((I-muIn)^2-sigmaIn^2)/(1+sigmaIn^2) - ((I-muOut)^2-sigmaOut^2)/(1+sigmaOut^2) 
     *                     + curv
     * assuming that the object and the background have a constant variance
     **/
    public void evolveVariance2() {
	double muI, muO, sigmaI, sigmaO;

	muO = meanOutside();
	muI = meanInside();
	sigmaO = varianceOutside(muO);
	sigmaI = varianceInside(muI);

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) {
		int I = getI(i,j);
		double curvTerm = getCurv(i,j);
		if( curvTerm < 0) curvTerm = 0.0;
		double crit = lambda1 * ((I-muI)*(I-muI) -sigmaI)/(1 + sigmaI) - lambda2 *  ((I-muO)*(I-muO) -sigmaO)/(1+sigmaO) + lambda3 * curvTerm;
			levelSet[i][j] += crit;
	    }
    }
  
    /**
     * evolution criterion : lambda1 * (k * curv - grad(k).norm ) norm
     * with k = 1 / ( 1 + |grad(I)| )
     **/
    public void evolveGradient1() {

	for(int xp = 1; xp < w-2; xp++)
	    for(int yp = 1; yp < h-2; yp++) {
		double Ix, Iy, Ixx, Iyy, Ixy;

		Ix = getI(xp+1, yp) - getI(xp, yp);
		Iy = getI(xp, yp+1) - getI(xp, yp);
		Ixx = getI(xp+2, yp) - 2* getI(xp+1, yp) + getI(xp, yp);
		Iyy = getI(xp, yp+2) - 2* getI(xp, yp+1) + getI(xp, yp);
		Ixy = getI(xp+1, yp+1)  - getI(xp+1, yp) - getI(xp, yp+1) + getI(xp, yp);

		double DI = Math.sqrt(Ix*Ix + Iy*Iy);
		double k = 1.0 / ( 1 + DI );

		// kx = -k^2/|gradI| (Ix*Ixx+Iy*Ixy)
		// ky = -k^2/|gradI| (Ix*Ixy+Iy*Iyy)

		double kx = 0.0, ky = 0.0, kN = 0.0;
		if( DI != 0) {
		    kx = - k*k*(Ix*Ixx+Iy*Ixy) / DI;
		    ky = - k*k*(Ix*Ixy+Iy*Iyy) / DI;
		    double [] norm = getNorm(xp, yp);
		    kN = norm[0]*kx + norm[1]*ky;
		}

		double crit = lambda1* (k * getCurv(xp,yp) - kN);
		if(Math.abs(crit) > Math.min(w,h)/3.0)   crit = 0;
		levelSet[xp][yp] += crit;
	    }
    }
    double getCurv(int x, int y) {
	if(( bi == null)||(x > w - 3)||(y > h-3))
	    return 0.0;

	double dx = levelSet[x+1][y] - levelSet[x][y];
	double dy = levelSet[x][y+1] - levelSet[x][y];
	double d2x = levelSet[x+2][y] - 2*levelSet[x+1][y] + levelSet[x][y];
	double d2y = levelSet[x][y+2] - 2*levelSet[x][y+1] + levelSet[x][y];
	double d2xy = levelSet[x+1][y+1] - levelSet[x+1][y] - levelSet[x][y+1] + levelSet[x][y];
	double norm = Math.sqrt(dx*dx+dy*dy);
	if(norm == 0) return 0.0; 
	return (d2x*dy*dy - 2*dx*dy*d2xy + d2y*dx*dx)/(norm*norm*norm);
    }

    double [] getNorm(int x, int y) {
	double [] res = new double[2];

	if(bi == null) return res;
	
	res[0] = levelSet[x+1][y] - levelSet[x][y];
	res[1] = levelSet[x][y+1] - levelSet[x][y];
	double denom = Math.sqrt(res[0]*res[0]+res[1]*res[1]);
	if(denom != 0) {
	    res[0] /= denom;
	    res[1] /= denom;
	}
	else {
	    res[0] = 0;
	    res[1] = 0;
	}
	return res;
    }

   /**
     * evolution criterion : lambda1 * (k * curv - grad(k).norm ) norm
     * with k = 1 / ( 1 + |grad(I)|^2 )
     **/
    public void evolveGradient2() {

	for(int xp = 0; xp < w-2; xp++)
	    for(int yp = 0; yp < h-2; yp++) {
		double Ix, Iy, Ixx, Iyy, Ixy;

		Ix = getI(xp+1, yp) - getI(xp, yp);
		Iy = getI(xp, yp+1) - getI(xp, yp);
		Ixx = getI(xp+2, yp) - 2* getI(xp+1, yp) + getI(xp, yp);
		Iyy = getI(xp, yp+2) - 2* getI(xp, yp+1) + getI(xp, yp);
		Ixy = getI(xp+1, yp+1)  - getI(xp+1, yp) - getI(xp, yp+1) + getI(xp, yp);

		double DI2 = Ix*Ix + Iy*Iy;
		double k = 1.0 / ( 1 + DI2 );

		double kx = 0.0, ky = 0.0, kN = 0.0;
		if( DI2 != 0) {
		    kx = - 2.0*k*k*(Ix*Ixx+Iy*Ixy);
		    ky = - 2.0*k*k*(Ix*Ixy+Iy*Iyy);
		    double [] norm = getNorm(xp, yp);
		    kN = norm[0]*kx + norm[1]*ky;
		}

		double crit = lambda1* (k * getCurv(xp,yp) - kN);
		if(Math.abs(crit) > Math.min(w,h)/3.0)   crit = 0;
		levelSet[xp][yp] += crit;
	    }
    }

   /**
     * evolution criterion : lambda1 * k * (lambda2 + curv) norm
     * with k = 1 / ( 1 + |grad(I)| )
     **/
    public void evolveGradient3() {
	double Ix, Iy;

	for(int xp = 0; xp < w-1; xp++)
	    for(int yp = 0; yp < h-1; yp++) {

		Ix = getI(xp+1, yp) - getI(xp, yp);
		Iy = getI(xp, yp+1) - getI(xp, yp);

		double DI = Math.sqrt(Ix*Ix + Iy*Iy);
		double k = 1.0 / ( 1 + DI );
		double crit = lambda1* k * (getCurv(xp,yp) + lambda2);
		if(Math.abs(crit) > Math.min(w,h)/3.0)   crit = 0;
		levelSet[xp][yp] += crit;
		
	    }
    }

   /**
     * evolution criterion : lambda1 * k * (lambda2 + curv) norm
     * with k = 1 / ( 1 + |grad(I)|^2 )
     **/
    public void evolveGradient4() {
	double Ix, Iy;
	
	for(int xp = 0; xp < w-1; xp++)
	    for(int yp = 0; yp < h-1; yp++) {
		Ix = getI(xp+1, yp) - getI(xp, yp);
		Iy = getI(xp, yp+1) - getI(xp, yp);

		double DI2 = Ix*Ix + Iy*Iy;
		double k = 1.0 / ( 1 + DI2 );
		double crit = lambda1* k * (getCurv(xp,yp) + lambda2);
		if(Math.abs(crit) > Math.min(w,h)/3.0)   crit = 0;
		levelSet[xp][yp] += crit;
	    }
    }

    /**
    * computes and returns the mean of the images points that are inside the BSpline of color color, 
    * using the mask m given as a parameter.
    * @param m mask already computed
    * @param color the color used for this region
    **/
    public double meanInside() {
	double mean = 0;
	int nbPoints = 0;

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) 
		if(levelSet[i][j] < 0) {
		    mean += getI(i,j);
		    nbPoints ++;
		}
	if(nbPoints == 0) return -1;
	mean /= nbPoints;
	System.out.println(" mean = " + mean + " with " + nbPoints + " points.");
	return mean;
    }
    /**
    * computes and returns the variance of the images points that are inside the BSpline of color color, using the mask m
    * given as a parameter and the mean mu.
    **/
     public double varianceInside(double mu) {
	double sigma = 0;
	int nbPoints = 0;

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) 
		if(levelSet[i][j] < 0) { // we are inside the region of color color
		    int col = getI(i,j);
		    sigma += (col - mu) * (col - mu);
		    nbPoints ++;
		}
	if(nbPoints == 0) return -1;
	sigma /= nbPoints;
	return sigma;
    }
    /**
    * computes and returns the mean of the images points that are outside the BSpline, using the mask m
    * given as a parameter.
    **/
    public double meanOutside() {
	double mean = 0;
	int nbPoints = 0;

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) 
		if(levelSet[i][j] > 0) { // we are outside
		    mean += getI(i,j);
		    nbPoints ++;
		}
	if(nbPoints == 0) return -1;
	mean /= nbPoints;
	return mean;
    }
   /**
    * computes and returns the variance of the images points that are outside the BSpline, using the mask m
    * given as a parameter and the mean mu.
    **/
     public double varianceOutside(double mu) {
	double sigma = 0;
	int nbPoints = 0;

	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) 
		if(levelSet[i][j] > 0) { // we are outside
		    int color = (getI(i,j));
		    sigma += (color - mu) * (color - mu);
		    nbPoints ++;
		}
	if(nbPoints == 0) return -1;
	sigma /= nbPoints;
	return sigma;
    }

    /**
     * used in public void regulDistances(int z)
     * test if 2 neighbours at the isocurve 0 have correct distances
     **/

    private void regulNeighbour(double [][] ls, int i0, int j0, int i1, int j1) {
	double d1 = Math.abs(ls[i0][j0])+Math.abs(ls[i1][j1]);
	if( ls[i0][j0]*ls[i1][j1] <= 0) {
	    if(d1 != 0) {
		ls[i0][j0] /= d1;
		ls[i1][j1] /= d1;
	    }
	    else if(ls[i0][j0] > 0){
		ls[i0][j0] = 0.5;
		ls[i1][j1] = -0.5;
	    }
	    else {
		ls[i0][j0] = - 0.5;
		ls[i1][j1] = 0.5;
	    }
	}
    }

    /** 
     * since the distance map is not a real distance map after evolution,
     * it is necessary to update it sometimes. new version of 25 fev with connexity = 8
     *
     * @param z not useful at all. was for BSPanel compatibility
     **/
    public void regulDistances(int z) {
	double [][] temp = new double[w][h];
	int max = Math.max(w,h);
	boolean [][] seen = new boolean[w][h];
	boolean [][] added = new boolean[w][h];

	double r2 = Math.sqrt(2);

	// we mark outside pixels by max and inside by -max
	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++)
		if(levelSet[i][j] > 0)		    temp[i][j] = max;
		else if(levelSet[i][j] < 0)	    temp[i][j] = -max;

	// we mark all points as non already seen and non already added in a list
	for(int i = 0; i < w; i++)
	    for(int j = 0; j < h; j++) {
		seen[i][j] = false;   added[i][j] = false;	    }

	// we now mark the list of inside and outside points
	ArrayList inside = new ArrayList();
	ArrayList outside = new ArrayList();

	for(int i = 1; i < w-1; i++)
	    for(int j = 1; j < h-1; j++) {
		if(levelSet[i][j] == 0) {
		    inside.add(new Point(i,j)); added[i][j] = true; seen[i][j] = true;  }
		// if negative and 1 neighbour positive
		else if( (levelSet[i][j] < 0) && ((levelSet[i+1][j]>0)||(levelSet[i][j+1]>0)||(levelSet[i-1][j]>0)||(levelSet[i][j-1]>0)) ) {
		    inside.add(new Point(i,j));   added[i][j] = true; seen[i][j] = true;  }
		// if positive and 1 neighbour negative or null
		else if( (levelSet[i][j] > 0) && ((levelSet[i+1][j]<=0)||(levelSet[i][j+1]<=0)||(levelSet[i-1][j]<=0)||(levelSet[i][j-1]<=0)) ) {
		    outside.add(new Point(i,j));   added[i][j] = true; seen[i][j] = true;  }
	    }

	for(int j = 1; j < h-1; j++) {
	    // case of i = 0
	    if(levelSet[0][j] == 0) {   inside.add(new Point(0,j));    added[0][j] = true;  seen[0][j] = true;  }
	    else if( (levelSet[0][j] < 0) && ((levelSet[1][j]>0)||(levelSet[0][j+1]>0)||(levelSet[0][j-1]>0)) ) {
		inside.add(new Point(0,j));   added[0][j] = true; seen[0][j] = true;  }
	    else if( (levelSet[0][j] > 0) && ((levelSet[1][j]<=0)||(levelSet[0][j+1]<=0)||(levelSet[0][j-1]<=0)) ) {
		    outside.add(new Point(0,j));   added[0][j] = true; seen[0][j] = true;  }
	    // case of i = w-1
	    if(levelSet[w-1][j] == 0) {   inside.add(new Point(w-1,j));    added[w-1][j] = true; seen[w-1][j] = true;  }
	    else if( (levelSet[w-1][j] < 0) && ((levelSet[w-1][j+1]>0)||(levelSet[w-2][j]>0)||(levelSet[w-1][j-1]>0)) ) {
		inside.add(new Point(w-1,j));   added[w-1][j] = true; seen[w-1][j] = true;  }
	    else if( (levelSet[w-1][j] > 0) && ((levelSet[w-1][j+1]<=0)||(levelSet[w-2][j]<=0)||(levelSet[w-1][j-1]<=0)) ) {
		outside.add(new Point(w-1,j));   added[w-1][j] = true; seen[w-1][j] = true;  }
	}
	// case of i = 0 and j = 0
	if(levelSet[0][0] == 0) {   inside.add(new Point(0,0));    added[0][0] = true; seen[0][0] = true;  }
	else if( (levelSet[0][0] < 0) && ((levelSet[1][0]>0)||(levelSet[0][1]>0)) ) {
	    inside.add(new Point(0,0));   added[0][0] = true; seen[0][0] = true;  }
	else if( (levelSet[0][0] > 0) && ((levelSet[1][0]<=0)||(levelSet[0][1]<=0)) ) {
	    outside.add(new Point(0,0));   added[0][0] = true; seen[0][0] = true;  }
	// case of i = 0 and j = h-1
	if(levelSet[0][h-1] == 0) {   inside.add(new Point(0,h-1));    added[0][h-1] = true; seen[0][h-1] = true;  }
	else if( (levelSet[0][h-1] < 0) && ((levelSet[1][h-1]>0)||(levelSet[0][h-2]>0)) ) {
	    inside.add(new Point(0,h-1));   added[0][h-1] = true; seen[0][h-1] = true;  }
	else if( (levelSet[0][h-1] > 0) && ((levelSet[1][h-1]<=0)||(levelSet[0][h-2]<=0)) ) {
	    outside.add(new Point(0,h-1));   added[0][h-1] = true; seen[0][h-1] = true;  }
	// case of i = w-1 and j = 0
	if(levelSet[w-1][0] == 0) {   inside.add(new Point(w-1,0));    added[w-1][0] = true; seen[w-1][0] = true;  }
	else if( (levelSet[w-1][0] < 0) && ((levelSet[w-1][1]>0)||(levelSet[w-2][0]>0)) ) {
	    inside.add(new Point(w-1,0));   added[w-1][0] = true; seen[w-1][0] = true;  }
	else if( (levelSet[w-1][0] > 0) && ((levelSet[w-1][1]<=0)||(levelSet[w-2][0]<=0)) ) {
	    outside.add(new Point(w-1,0));   added[w-1][0] = true; seen[w-1][0] = true;  }
	// case of i = w-1 and j = h-1
	if(levelSet[w-1][h-1] == 0) {   inside.add(new Point(w-1,h-1));    added[w-1][h-1] = true; seen[w-1][h-1] = true;  }
	else if( (levelSet[w-1][h-1] < 0) && ((levelSet[w-2][h-1]>0)||(levelSet[w-1][h-2]>0)) ) {
	    inside.add(new Point(w-1,h-1));   added[w-1][h-1] = true; seen[w-1][h-1] = true;  }
	else if( (levelSet[w-1][h-1] > 0) && ((levelSet[w-2][h-1]<=0)||(levelSet[w-1][h-2]<=0)) ) {
	    outside.add(new Point(w-1,h-1));   added[w-1][h-1] = true; seen[w-1][h-1] = true;  }
	// case of j = 0 and 0<i<w-1
	for(int i = 1; i < w-1; i++) {
	    if(levelSet[i][0] == 0) {
		inside.add(new Point(i,0));    added[i][0] = true; seen[i][0] = true;  }
	    else if( (levelSet[i][0] < 0) && ((levelSet[i+1][0]>0)||(levelSet[i][1]>0)||(levelSet[i-1][0]>0)) ) {
		inside.add(new Point(i,0));   added[i][0] = true;seen[i][0] = true;  }
	    else if( (levelSet[i][0] > 0) && ((levelSet[i+1][0]<=0)||(levelSet[i][1]<=0)||(levelSet[i-1][0]<=0)) ) {
		outside.add(new Point(i,0));   added[i][0] = true;seen[i][0] = true;  }
	}
	// case of j = h-1 and 0<i<w-1
	for(int i = 1; i < w-1; i++) {
	    if(levelSet[i][h-1] == 0) {
		inside.add(new Point(i,h-1));    added[i][h-1] = true; seen[i][h-1] = true;  }
	    else if( (levelSet[i][h-1] < 0) && ((levelSet[i+1][h-1]>0)||(levelSet[i-1][h-1]>0)||(levelSet[i][h-2]>0)) ) {
		inside.add(new Point(i,h-1));   added[i][h-1] = true; seen[i][h-1] = true;  }
	    else if( (levelSet[i][h-1] > 0) && ((levelSet[i+1][h-1]<=0)||(levelSet[i-1][h-1]<=0)||(levelSet[i][h-2]<=0)) ) {
		outside.add(new Point(i,h-1));   added[i][h-1] = true; seen[i][h-1] = true;  }
	}
	
	// inside initialisation
	for(int a = 0; a < inside.size(); a++) {
	    Point p = (Point) inside.get(a);
	    int i = (int) p.getX(); int j = (int)  p.getY();
	    temp[i][j] = Math.max(levelSet[i][j],-0.5);
	}
	//outside initialisation
	for(int a = 0; a < outside.size(); a++) {
	    Point p = (Point) outside.get(a);
	    int i = (int)  p.getX(); int j = (int)  p.getY();
	    temp[i][j] = Math.min(levelSet[i][j],0.5);
	}
	
	levelSet=temp;

	// inside propagation
	for(int a = 0; a < inside.size(); a++) {
	    Point p = (Point) inside.get(a);
	    int i = (int)  p.getX(); int j = (int)  p.getY();
	    seen[i][j]=true;
	    double dLocal = levelSet[i][j];
	    // point [i-1][j]
	    if( (i>0) && !(seen[i-1][j]) && (levelSet[i-1][j]<dLocal-1)) {
		levelSet[i-1][j] = dLocal-1;    
		if(!added[i-1][j]) { inside.add(inside.size(),new Point(i-1,j));  added[i-1][j]=true;}
	    }
	    // point [i+1][j]
	    if( (i<w-1) && !(seen[i+1][j]) && (levelSet[i+1][j]<dLocal-1)) {
		levelSet[i+1][j] = dLocal-1;   
		if(!added[i+1][j]) { inside.add(inside.size(),new Point(i+1,j));   added[i+1][j]=true;}
	    }
	    // point [i][j-1]
	    if( (j>0) && !(seen[i][j-1]) && (levelSet[i][j-1]<dLocal-1)) {
		levelSet[i][j-1] = dLocal-1;    
		if(!added[i][j-1]) { inside.add(inside.size(),new Point(i,j-1));   added[i][j-1]=true;}
	    }
	    // point [i][j+1]
	    if( (j<h-1) && !(seen[i][j+1]) && (levelSet[i][j+1]<dLocal-1)) {
		levelSet[i][j+1] = dLocal-1;    
		if(!added[i][j+1]) { inside.add(inside.size(),new Point(i,j+1));   added[i][j+1]=true;}
	    }
	    // point [i+1][j+1]
	    if( (i<w-1) && (j<h-1) && !(seen[i+1][j+1])&& (levelSet[i+1][j+1]<dLocal-r2)) {
		levelSet[i+1][j+1] = dLocal-r2;    
		if(!added[i+1][j+1]) { inside.add(inside.size(),new Point(i+1,j+1));   added[i+1][j+1]=true;}
	    }
	    // point [i+1][j-1]
	    if( (i<w-1) && (j>0) && !(seen[i+1][j-1])&& (levelSet[i+1][j-1]<dLocal-r2)) {
		levelSet[i+1][j-1] = dLocal-r2;    
		if(!added[i+1][j-1]) { inside.add(inside.size(),new Point(i+1,j-1));   added[i+1][j-1]=true;}
	    }
	    // point [i-1][j+1]
	    if( (i>0) && (j<h-1)  && !(seen[i-1][j+1])&& (levelSet[i-1][j+1]<dLocal-r2)) {
		levelSet[i-1][j+1] = dLocal-r2;    
		if(!added[i-1][j+1]) { inside.add(inside.size(),new Point(i-1,j+1));   added[i-1][j+1]=true;}
	    }
	    // point [i-1][j-1]
	    if( (i>0) && (j>0) && !(seen[i-1][j-1])&& (levelSet[i-1][j-1]<dLocal-r2)) {
		levelSet[i-1][j-1] = dLocal-r2;    
		if(!added[i-1][j-1]) { inside.add(inside.size(),new Point(i-1,j-1));   added[i-1][j-1]=true;}
	    }
	    inside.remove(a);
	    a--;
	}

	// outside propagation
	for(int a = 0; a < outside.size(); a++) {
	    Point p = (Point) outside.get(a);
	    int i = (int)  p.getX(); int j =  (int) p.getY();
	    seen[i][j]=true;
	    double dLocal = levelSet[i][j];
	    if( (i>0) && !(seen[i-1][j]) && (levelSet[i-1][j]>dLocal+1)) {
		levelSet[i-1][j] = dLocal+1;    
		if(!added[i-1][j]) { outside.add(outside.size(),new Point(i-1,j));  added[i-1][j]=true;}
	    }
	    if( (i<w-1) && !(seen[i+1][j]) && (levelSet[i+1][j]>dLocal+1)) {
		levelSet[i+1][j] = dLocal+1;   
		if(!added[i+1][j]) { outside.add(outside.size(),new Point(i+1,j));   added[i+1][j]=true;}
	    }
	    if( (j>0) && !(seen[i][j-1]) && (levelSet[i][j-1]>dLocal+1)) {
		levelSet[i][j-1] = dLocal+1;    
		if(!added[i][j-1]) { outside.add(outside.size(),new Point(i,j-1));   added[i][j-1]=true;}
	    }
	    if( (j<h-1) && !(seen[i][j+1]) && (levelSet[i][j+1]>dLocal+1)) {
		levelSet[i][j+1] = dLocal+1;    
		if(!added[i][j+1]) { outside.add(outside.size(),new Point(i,j+1));   added[i][j+1]=true;}
	    }
	    // point [i+1][j+1]
	    if( (i<w-1) && (j<h-1) && !(seen[i+1][j+1]) && (levelSet[i+1][j+1]>dLocal+r2)) {
		levelSet[i+1][j+1] = dLocal+r2;    
		if(!added[i+1][j+1]) { outside.add(outside.size(),new Point(i+1,j+1));   added[i+1][j+1]=true;}
	    }
	    // point [i+1][j-1]
	    if( (i<w-1) && (j>0)&& !(seen[i+1][j-1]) && (levelSet[i+1][j-1]>dLocal+r2)) {
		levelSet[i+1][j-1] = dLocal+r2;    
		if(!added[i+1][j-1]) { outside.add(outside.size(),new Point(i+1,j-1));   added[i+1][j-1]=true;}
	    }
	    // point [i-1][j+1]
	    if( (i>0) && (j<h-1) && !(seen[i-1][j+1]) && (levelSet[i-1][j+1]>dLocal+r2)) {
		levelSet[i-1][j+1] = dLocal+r2;    
		if(!added[i-1][j+1]) { outside.add(outside.size(),new Point(i-1,j+1));   added[i-1][j+1]=true;}
	    }
	    // point [i-1][j-1]
	    if( (i>0) && (j>0)&& !(seen[i-1][j-1]) && (levelSet[i-1][j-1]>dLocal+r2)) {
		levelSet[i-1][j-1] = dLocal+r2;    
		if(!added[i-1][j-1]) { outside.add(outside.size(),new Point(i-1,j-1));   added[i-1][j-1]=true;}
	    }
	    outside.remove(a);
	    a--;
	}

	// writing the BufferedImage
	if(capture) {
	    BufferedImage bd = new BufferedImage(w,h,BufferedImage.TYPE_INT_RGB);
	    int col;

	    for(int i = 0; i < w; i++)
		for(int j = 0 ; j < h ; j++) {
		    if(levelSet[i][j] <0) {
			col = Math.min(255,(int) -levelSet[i][j]);
			col *= 257;
		    }
		    else {
			col = Math.min(255,(int) (levelSet[i][j]));
			col *= 65537;
		    }
		    bd.setRGB(i,j,col);
		}
	    try{
		if(iter<10)
		ImageIO.write(bd,"png",new File("regul0"+ iter +".png"));
		else
		ImageIO.write(bd,"png",new File("regul"+ iter +".png"));
	    }
	    catch(Exception e) {System.out.println("pas de regul bd");}
	    // draw and save the image with current isolevel 0
	    Graphics g = bd.getGraphics();
	    paintComponent(g);
	    try{
			if(iter<10)
			    ImageIO.write(bd,"png",new File("levelset0"+ iter +".png"));
			else
			    ImageIO.write(bd,"png",new File("levelset"+ iter +".png"));
	    }
	    catch(Exception e) {System.out.println("pas de regul bd");}
	    iter++;
	}
    }

    /**
     * paints this graphical component and the BSpline corresponding
     **/
    public void paintComponent(Graphics gc) {
	gc.setColor(Color.WHITE);
        super.paintComponent(gc);
	Graphics2D g2 = (Graphics2D) gc;
	
	if(distMapShow) {
	    int max = Math.max(w,h);
	    for(int i = 1; i < w-1; i++)
		for(int j = 1; j < h-1; j++) {
		    if(levelSet[i][j] > 0) {
			int col = Math.max(0,Math.min(255,(int) (255.0*2*levelSet[i][j]/(max+1))));
			gc.setColor(new Color(col,0,col));
		    }
		    else {
			int col = Math.max(0,Math.min(255,(int) (-255.0*2*levelSet[i][j]/(max+1))));
			gc.setColor(new Color(0,col,col));
		    }
		    gc.fillRect(i,j,1,1);
		    // draw the isocontour 0
		    if((levelSet[i][j]*levelSet[i-1][j] <= 0) || (levelSet[i][j]*levelSet[i][j-1] <= 0)) {
			gc.setColor(Color.WHITE);
			gc.fillRect(i-1,j-1,3,3);
		    }
		}
	}
	else {
	    if(bi != null) {
		g2.drawImage(bi,null,0,0);

		if( (w > levelSet.length) || ( h > levelSet[0].length)) {
		    int max = Math.max(w,h);
		    double [][] tmp = new double[w][h];
		    for(int i = 0; i < w; i++)
			for(int j = 0; j < h; j++ )
			    tmp[i][j] = max;
		    for(int i = 0; i < Math.min(levelSet.length,w); i++)
			for(int j = 0; j < Math.min(levelSet[0].length,h); j++ )
			    tmp[i][j] = levelSet[i][j];

		    levelSet = tmp;
		}
		    
		if(withMask) {
		    gc.setColor(Color.RED);
		    for(int i = 0; i < w; i++)
			for(int j = 0; j < h; j++)
			    if(levelSet[i][j] < 0) 
				gc.fillRect(i,j,1,1);
		}

		for(int i = 1; i < w-1; i++)
		    for(int j = 1; j < h-1; j++)

			if(levelSet[i][j] <= 0) {
			    if ( (levelSet[i][j-1]>=0)||(levelSet[i-1][j-1]>=0)||(levelSet[i-1][j]>=0)||
				 (levelSet[i][j+1]>=0)||(levelSet[i+1][j+1]>=0)||(levelSet[i+1][j]>=0)||
				 (levelSet[i-1][j+1]>=0) || (levelSet[i+1][j-1]>=0) ) {
				
				gc.setColor(Color.GREEN);
				gc.fillOval(i-2,j-2,4,4);
				//double [] norm = getNorm(i,j);
				//gc.setColor(Color.BLUE);
				//gc.drawLine((int) (i+norm[0]),(int) (j+norm[1]),(int) (i+norm[0]*10),(int) (j+norm[1]*10));
			    }
			}
	    }
	    
	}
    }
    /**
     * background is supposed to be 0
     * @param r is assumed to be of dimension (w;h)
     **/
    public double [][] thin(double [][] ro) {
        int nbVoisins;
        int nbTransitions;
	int max = Math.max(w,h);

        // initialisations
        int [][] g = new int[w][h];
        double [][] r = new double[w][h];

        boolean [] p = new boolean[8];
        for(int i=0; i<w; i++) {
            for(int j=0; j<h; j++) {
                g[i][j] = 0;
		r[i][j] = ro[i][j];
            }
	}
	
        // 1ere passe
        for(int i=1; i<w-1; i++) {
            for(int j=1; j<h-1; j++) {
                if(r[i][j] == -1) { // si on est dans la region qui nous interesse
                    // calcul du nombre de voisins
                    p[0] = (r[i][j+1] ==-1);
                    p[1] = (r[i+1][j+1] ==-1);
                    p[2] = (r[i+1][j] ==-1);
                    p[3] = (r[i+1][j-1] ==-1);
                    p[4] = (r[i][j-1] ==-1);
                    p[5] = (r[i-1][j-1] ==-1);
                    p[6] = (r[i-1][j] ==-1);
                    p[7] = (r[i-1][j+1] ==-1);
                    nbVoisins = 0;
                    for(int k=0; k<8; k++)
                        if(p[k]) nbVoisins ++;
                    if(nbVoisins>= 2 && nbVoisins<=6) {
                        // calcul du nombre de transitions
			nbTransitions = 0;
                        for(int k=0; k<7; k++)
                            if(!p[k]&&p[k+1]) nbTransitions++;
                        if(!p[7]&&p[0]) nbTransitions++;
			//System.out.print("[");
			//for(int k=0; k<8; k++) System.out.print(p[k]+" ");
			//System.out.println("");
			if(nbTransitions==1)
                            if(!p[0]||!p[2]||!p[4])
                                if(!p[2]||!p[4]||!p[6]) 
                                    g[i][j] = 255;
		    }
                }
            }
        }
        for(int i=1; i<w-1; i++) {
            for(int j=1; j<h-1; j++) {
                if(g[i][j] != 0) {
                    r[i][j] = max;
                    g[i][j] = 0;
                }
            }
        }

        // 2eme passe
        for(int i=1; i<w-1; i++) {
            for(int j=1; j<h-1; j++) {
                if(r[i][j] == -1) {
                    // calcul du nombre de voisins
                    p[0] = (r[i][j+1] ==-1);
                    p[1] = (r[i+1][j+1] ==-1);
                    p[2] = (r[i+1][j] ==-1);
                    p[3] = (r[i+1][j-1] ==-1);
                    p[4] = (r[i][j-1] ==-1);
                    p[5] = (r[i-1][j-1] ==-1);
                    p[6] = (r[i-1][j] ==-1);
                    p[7] = (r[i-1][j+1] ==-1);
                    nbVoisins = 0;
                    for(int k=0; k<8; k++)
                        if(p[k]) nbVoisins ++;
                    if(nbVoisins>= 2 && nbVoisins<=6) {
                        // calcul du nombre de transitions
                        nbTransitions = 0;
                        for(int k=0; k<7; k++)
                           if(!p[k]&&p[k+1]) nbTransitions++;
                        if(!p[7]&&p[0]) nbTransitions++;
			//System.out.print("]");
			//for(int k=0; k<8; k++) System.out.print(p[k]+" ");
			//System.out.println("");
                        if(nbTransitions==1) 
                            if(!p[0]||!p[2]||!p[6])
                                if(!p[0]||!p[4]||!p[6])
                                    g[i][j] = 255;
                    }
                }
            }
        }
        for(int i=1; i<w-1; i++) 
            for(int j=1; j<h-1; j++)
                if(g[i][j] != 0) {
                    r[i][j] = max;
                    g[i][j] = 0;
                }
	    
        return r;
    }

    // we assume that dim of tab are (w,h)
    public void skelett(double [][] tab) {
	boolean modif = false;
	double [][] tmp = new double[w][h];

	do{
	    modif = false;
	    tmp = thin(tab);
	    for(int i=0; i<w; i++)
		for(int j=0; j<h; j++)
		    if(tmp[i][j] != tab[i][j]) {
			tab[i][j] = tmp[i][j];
			modif = true;
		    }
	    
	} while(modif);
	System.out.println("");
    }

    public void checkOverlapping() {}
    public void cutBSplines() {}
    public void reorientCurv() {}
    public void loadBSplines(BufferedReader reader) {}
    public void saveBSplines(File file) {}
    public String saveBSplinesAsString() {return "";}
    public BSparam loadParameters(BufferedReader reader) {return null;}
    public void setDistant(int d) {}
    public void setClose(int c) {}
}
